# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs::Type::JIRA;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Extension::ExternalBugs::Type);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Extension::ExternalBugs::Util qw(mail_error);
use Bugzilla::Util qw(correct_urlbase);

use Data::Dumper;
use JSON;
use URI;
use XMLRPC::Lite;
use LWP::UserAgent;
use MIME::Base64;
use HTTP::Request;
use List::MoreUtils qw(any);

use constant TYPE => 'JIRA';

my $logger = Bugzilla->logger;

sub get_data {
  my ($self, $ext_bug) = @_;
  my $bug_data = {};

  # Do the rest call
  my $details = $self->_do_jira_rest_call($ext_bug->ext_bz_bug_id,
    "issue/" . $ext_bug->ext_bz_bug_id . "?fields=priority,status,summary", 'GET',);

  # Get the information we require from the result
  if ($details->{fields}{status}{name}) {
    $bug_data->{status} = $details->{fields}{status}{name};
  }
  if ($details->{fields}{priority}{name}) {
    $bug_data->{priority} = $details->{fields}{priority}{name};
  }
  if ($details->{fields}{summary}) {
    $bug_data->{description} = $details->{fields}{summary};
  }

  return $bug_data;
}

sub can_sync {
  my ($self, $ext_bug) = @_;
  return 1;
}

sub activiate {
  my ($self, $ext_bug) = @_;

  return () unless ($self->can_get());

  # If the status is still 'None', then it means we could need get the bug
  #  details. This could be because of an RPC failure, or because the bug
  #  doesn't exist. Let's assume it is the second (bug 720778)
  if ($ext_bug->ext_status eq 'None') {
    ThrowUserError('ext_bz_invalid_ext_bug_id',
      {badvalue => $ext_bug->ext_bz_bug_id});
  }

  # If the status is Closed, we won't be able to request updates for
  #  this issue
  elsif ($ext_bug->ext_status =~ m/^(:?Closed|Resolved)$/) {
    $logger->debug("JIRA XMLRPC: Cannot tag "
        . $ext_bug->ext_bz_bug_id
        . " for updates as it is closed");
    return 0;
  }

  # We need to mark this bug in JIRA so that it sends us updates
  return $self->set_jira_flags($ext_bug);
}

sub deactivate {
  my ($self, $ext_bug) = @_;

  # If the status is Closed, we won't be able to request updates for
  #  this issue
  if ($ext_bug->ext_status =~ m/^(:?Closed|Resolved)$/) {
    $logger->debug("JIRA XMLRPC: Cannot detag "
        . $ext_bug->ext_bz_bug_id
        . " for updates as it is closed");
    return 0;
  }

  # We need to mark this bug in JIRA so that it stops sending us updates
  return $self->set_jira_flags($ext_bug);
}

sub send_changes {
  my ($self, $params) = @_;

  my $ext_bug  = $params->{'ext_bug'};
  my $bug      = $params->{'bug'};
  my $diffs    = $params->{'diffs'};
  my $start    = $params->{'start'};
  my $end      = $params->{'end'};
  my $changer  = $params->{'changer'};
  my $comments = $params->{'comments'};
  my $private  = scalar(@{$bug->groups_in});

  my $group;

  if ($private) {
    if ($self->group_map && @{$self->group_map}) {
      foreach my $grp (@{$self->group_map}) {
        my $bzgrp = $grp->[0];
        if (any { $_->name eq $bzgrp } (@{$bug->groups_in})) {
          $group = $grp->[1];
          last;
        }
      }
    }

    # bail for private bugs with no matching group
    if (!$group) {
      Bugzilla->logger->info(
        "JIRA: No match for any group in group match for bug " . $bug->id);
      return;
    }
  }

  # Create a link to this bug using the format that JIRA likes
  my $buglink
    = "[bug "
    . $bug->id . '|'
    . correct_urlbase()
    . "show_bug.cgi?id="
    . $bug->id . ']';

  # Record whether ALL changes were successful
  my $ok = undef;

  # Record any change in status
  if ($self->post_status) {
    foreach my $diff (@$diffs) {
      my $who = $diff->{who};

      if ( $self->ignore_accounts
        && @{$self->ignore_accounts}
        && any { $who->login eq $_ } @{$self->ignore_accounts})
      {

        # Ignore changes from JIRA
        next;
      }

      my $what = $diff->{field_name};
      my $when = $diff->{bug_when};
      my $old  = $diff->{old};
      my $new  = $diff->{new};
      if ($what eq 'bug_status' and $old) {
        my $author = $who->identity;
        my $text   = "$author changed the Status of $buglink from $old to $new";
        my $result = $self->send_as_comment($ext_bug, $text, $private, $group);
        if (not defined $ok or $result == 0) {
          $ok = $result;
        }
      }
    }
  }

  # Record any change in comments
  if ($self->post_comments) {
    foreach my $comment (@$comments) {
      if (
           $self->ignore_accounts
        && @{$self->ignore_accounts}
        && any { $comment->author->login eq $_ }
        @{$self->ignore_accounts}
        )
      {

        # Ignore comments from JIRA
        next;
      }

      my $author = $comment->author->identity;
      my $text   = "$author made a comment on $buglink\n\n" . $comment->body;
      my $pc     = $private || $comment->is_private;
      my $comment_group;
      if ($comment->is_private) {
        if ($self->group_map && @{$self->group_map}) {
          foreach my $grp (@{$self->group_map}) {
            my $bzgrp = $grp->[0];
            if (any { $_->name eq Bugzilla->params->{insidergroup} } (@{$bug->groups_in})) {
              $comment_group = $grp->[1];
              last;
            }
          }
        }

        # bail for private comments with no map to the insider group
        if (!$comment_group) {
          Bugzilla->logger->info(
            "JIRA: No match for insider group in group match for private comment on bug "
              . $bug->id);
          next;
        }
      }

      my $result = $self->send_as_comment($ext_bug, $text, $pc, $comment_group);
      if (not defined $ok or $result == 0) {
        $ok = $result;
      }
    }
  }

  return $ok || 0;
}

sub send_as_comment {
  my ($self, $ext_bug, $comment, $is_private, $group) = @_;

  Bugzilla->job_queue->insert(
    'send_jira',
    {
      action        => 'send_as_comment',
      ext_bz_id     => $ext_bug->ext_bz_id,
      ext_bz_bug_id => $ext_bug->ext_bz_bug_id,
      comment       => $comment,
      is_private    => $is_private,
      group         => $group,
    }
  );

  return 1;
}

sub _send_as_comment {
  my ($self, $ext_bz_bug_id, $comment, $is_private, $group) = @_;

  my $details = {body => $comment};

  ## RED HAT EXTENSION START 826699
  # If it is a private bug (or private comment on a public bug), then
  # we need to restrict visibility on it.
  if ($is_private && $group) {
    $details->{visibility} = {type => 'group', value => $group,};
  }
  elsif ($is_private) {
    Bugzilla->logger->warning(
      "JIRA Attempt to send private comment to an unprotected tracker for $ext_bz_bug_id"
    );
    return 0;
  }
  ## RED HAT EXTENSION END 826699

  # We are going to assume that the JIRA issues needs not to be escaped
  # since it only contains alphanumeric characters and hyphens
  return $self->_do_jira_rest_call($ext_bz_bug_id, "issue/$ext_bz_bug_id/comment",
    'POST', $details,);
}

sub set_jira_flags {
  my ($self, $ext_bug) = @_;

  Bugzilla->job_queue->insert(
    'send_jira',
    {
      action        => 'set_jira_flags',
      ext_bz_id     => $ext_bug->ext_bz_id,
      ext_bz_bug_id => $ext_bug->ext_bz_bug_id,
    }
  );

  return 1;
}

sub _set_jira_flags {
  my ($self, $ext_bz_bug_id) = @_;
  my $return = 0;

  # Get the flags we need to update. Exit if neither is set
  my $push_cf_id     = $self->push_cf_id;
  my $backlink_cf_id = $self->backlink_cf_id;
  unless ($push_cf_id or $backlink_cf_id) {
    return;
  }

  # Get a list of bugs that have this JIRA issue
  my $sql = "
        SELECT bug_id
        FROM ext_bz_bug_map
        WHERE ext_bz_id = ?
            AND ext_bz_bug_id = ?
    ";

  my $bug_list
    = Bugzilla->dbh->selectcol_arrayref($sql, undef, $self->id, $ext_bz_bug_id);


  my $details = {};

  # Set the must-update-bugzilla custom field
  if ($push_cf_id) {

    # Do we need to update this bug list anymore
    my $do_update = scalar(@$bug_list) ? [{value => 'Perform'}] : [];

    $details->{update}{"customfield_$push_cf_id"} = [{set => $do_update}];
  }

  # Set the list of bugs that this JIRA is tracking
  if ($backlink_cf_id) {

    # Turn them into URLs
    my @bug_urls = map { correct_urlbase() . "show_bug.cgi?id=$_" } @$bug_list;

    $details->{update}{"customfield_$backlink_cf_id"}
      = [{set => join(",", @bug_urls)}];
  }

  if (scalar(keys(%$details)) == 0) {

    # There is nothing to update, so return
    return 1;
  }

  # We are going to assume that the JIRA issues needs not to be escaped
  # since it only contains alphanumeric characters and hyphens
  my $res
    = $self->_do_jira_rest_call($ext_bz_bug_id, "issue/$ext_bz_bug_id", 'PUT',
    $details,);

  if (not defined $res) {

    # We were returned with undef. This means we have failure
    return 0;
  }

  # We received a successful response
  return 1;
}

sub _do_jira_rest_call {
  my $self          = shift;
  my $ext_bz_bug_id = shift;
  my $call          = shift;
  my $method        = shift;
  my $callobj       = shift;

  my $auth;

  if ($self->ext_login && $self->ext_password) {
    $auth
      = MIME::Base64::encode_base64($self->ext_login . ':' . $self->ext_password);
  }
  else {
    $logger->debug(
      "JIRA XMLRPC Warning: No username/password information for " . $self->url);
  }


  # Send the request
  my $uri = URI->new($self->url . 'rest/api/2/' . $call);
  my $ua  = LWP::UserAgent->new;
  $ua->agent("RedHatBugzilla/" . BUGZILLA_VERSION);    # :)

  if ( Bugzilla->params->{external_tracker_proxy}
    && Bugzilla->params->{external_tracker_proxy} ne '')
  {
    $ua->proxy(['http', 'https'], Bugzilla->params->{external_tracker_proxy});
  }

  my $result  = undef;
  my $content = defined $callobj ? encode_json($callobj) : '';

  eval {
    # Create a request
    my @headers = ('content-type' => 'application/json;charset=UTF-8');
    push(@headers, ('authorization' => "Basic $auth")) if ($auth);

    my $req = HTTP::Request->new($method, $uri, \@headers, $content,);

    # Pass request to the user agent and get a response back
    my $res = $ua->request($req);

    $logger->debug("JIRA REST Calling: " . $uri->as_string);

    unless ($res->is_success) {
      die $res->status_line;
    }

    # Some calls return no content
    $result = $res->content ? decode_json($res->content) : '';
  };

  if ($@) {
    my $err = $@;
    ThrowUserError('ext_bz_jira_error', {errtext => "$err"});
  }

  if (ref $result and $result->{errors}) {
    my $error = Dumper($result->{errors});
    ThrowUserError('ext_bz_jira_error', {errtext => "$error"});
  }

  # Log a comment, except if we were trying to log in
  $logger->debug(
    "JIRA REST Success: $call call made for JIRA issue $ext_bz_bug_id");

  return $result;
}

1;

__END__

=head1 NAME

Bugzilla::Extension::ExternalBugs::Type::JIRA

=head1 DESCRIPTION

Type represents a JIRA issue object. It is an implementation of
L<Bugzilla::Extension::ExternalBugs::Type>, and thus provides all methods that
L<Bugzilla::Extension::ExternalBugs::Type> provides.

This type can sync data from remote JIRA sites to the local site.

This type can push bug changes as comments to remote JIRA sites.

B<BUGBUG> This type is unfortunately tied to JBoss Jira and requires refactoring
to be more generally useful. See L<https://bugzilla.redhat.com/show_bug.cgi?id=1290545>

=cut
