# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs::Type::SFDC_Sandbox;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Extension::ExternalBugs::Type);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::Error;

use Bugzilla::WebService::Server::XMLRPC;
use Bugzilla::WebService::Bug;

use Bugzilla::Extension::ExternalBugs::Util qw(mail_error sfdc_get_directory);

use Data::Dumper;
use File::Spec::Functions qw(catfile);
use SOAP::Lite;
use WWW::Salesforce::Simple;

use constant TYPE => 'SFDC_Sandbox';

our $WSDL_XML = join('', <DATA>);
our $HOSTNAME = Bugzilla->params->{'sfdc_sandbox_server'};
our $SERVER   = "https://test.salesforce.com/services/Soap/u/19.0";
our $CACHE    = sfdc_get_directory('Sandbox', $HOSTNAME, $WSDL_XML);
our $WSDL     = 'file:' . catfile($CACHE, 'BugWebServicesSandbox.wsdl.xml');

$WWW::Salesforce::SF_URI = 'urn:partner.soap.sforce.com';

my $logger = Bugzilla->logger;

sub get_data {
  my ($self, $ext_bug) = @_;

  unless ($ext_bug->ext_bz_bug_id =~ /^\d+$/) {
    $logger->debug("SFDC Bug id " . $ext_bug->ext_bz_bug_id . " is not a number.");
    return;
  }

  unless ($self->ext_login && $self->ext_password) {
    $logger->debug(
      "SFDC SOAP: Warning: No username/password information for " . $self->url);
    return;
  }

  my $result = undef;
  eval {
    my $sforce = WWW::Salesforce::Simple->new(
      serverurl => $SERVER,
      username  => $self->ext_login,
      password  => $self->ext_password,
    );
    if (not $sforce) {
      die "SFDC SOAP: WWW::Salesforce login failure";
    }
    $logger->debug("SFDC SOAP: Login successful");

    my $soql
      = "SELECT Severity__c, Status, Subject, Account.Name FROM Case WHERE CaseNumber = '"
      . $ext_bug->ext_bz_bug_id . "'";

    $result = $sforce->do_query($soql);
  };

  if (my $call_error = $@) {
    ThrowCodeError('eval_error', {command => "SFDC SOAP", msg => $@});
  }

  if (!$result) {

    # Probably not a valid case number
    return;
  }

  my %data
    = (priority => $result->[0]{Severity__c}, status => $result->[0]{Status},);

  if ( defined($result->[0]{'Account'})
    && defined($result->[0]{'Account'}->{'Name'}))
  {
    $data{description}
      = '[' . $result->[0]{'Account'}->{'Name'} . '] ' . $result->[0]{Subject};
  }
  else {
    $data{description} = '[unknown] ' . $result->[0]{Subject};
  }

  return \%data;
}

sub can_sync {
  my ($self, $ext_bug) = @_;
  return 1;
}

sub send_changes {
  my ($self, $params) = @_;

  $logger->debug("SFDC SOAP: Entering send_changes");

  my $ext_bug = $params->{'ext_bug'};
  my $bug     = $params->{'bug'};
  my $start   = $params->{'start'};

  my $serializer = Bugzilla::XMLRPC::Serializer->new();

  # Send a full copy of the bug report using Bug.get
  my $bug_data = Bugzilla::WebService::Bug->get({
    ids          => [$bug->id],
    extra_fields => ['external_bugs', 'description', 'flags'],
  });
  my $bug_xml_data = $serializer->envelope('response', 'Bug.get', $bug_data);

  # Get the comments for this bug
  my $bug_comments = Bugzilla::WebService::Bug->comments(
    {ids => [$bug->id], new_since => $start});
  my $bug_xml_comments
    = $serializer->envelope('response', 'Bug.comments', $bug_comments);

  ## REDHAT EXTENSION 965415: Bugzilla 4.4 switches the above calls to
  # the shadow db. You need to switch make to the main one
  Bugzilla->switch_to_main_db();

  # We now need to send the events to the Job Queue. However, we must set
  # We need to record the ext_bz info since there is no guarantee that the
  # row will exist at the time we send the event.
  Bugzilla->job_queue->insert(
    'send_sfdc',
    {
      ext_bz_type_id   => $ext_bug->ext_bz_id,
      bug_id           => $ext_bug->bug_id,
      bug_xml_data     => $bug_xml_data,
      bug_xml_comments => $bug_xml_comments,
    }
  );

  return 1;
}

sub send_to_sfdc {
  my $self             = shift;
  my $bug_id           = shift;
  my $bug_xml_data     = shift;
  my $bug_xml_comments = shift;

  unless ($self->ext_login && $self->ext_password) {
    $logger->debug(
      "SFDC SOAP: Warning: No username/password information for " . $self->url);
    die "No user name and password supplied";
  }

  eval {
    my $sforce = WWW::Salesforce::Simple->new(
      serverurl => $SERVER,
      username  => $self->ext_login,
      password  => $self->ext_password,
    );
    if (not $sforce) {
      die "SFDC SOAP: WWW::Salesforce login failure";
    }
    $logger->debug("SFDC SOAP: Login successful");
    my $client = $sforce->get_client(1);
    my @errors;
    my $result = $client->cache_dir($CACHE)->service($WSDL)
      ->updateBug($bug_xml_data, $sforce->get_session_header());
    push(@errors, "updateBug returned failure ($result)")
      if (!$result || $result eq 'false');
    if ($bug_xml_comments) {
      $result = $client->cache_dir($CACHE)->service($WSDL)
        ->upsertComments($bug_xml_comments, $sforce->get_session_header());
      push(@errors, "upsertComments returned failure ($result)")
        if (!$result || $result eq 'false');
    }

    if (@errors) {
      die("Call(s) failed: " . join(", ", @errors));
    }
  };


  if (my $call_error = $@) {
    die $call_error;
  }

  $logger->debug("SFDC SOAP: Success: changes sent for bug $bug_id");
  return 1;
}

sub sfdc_hostname {
  return 'c.cs11.visual.force.com';
}

1;

__DATA__
<?xml version="1.0" encoding="UTF-8"?>
<!--
 Web Services API : BugWebServices
-->
<definitions targetNamespace="http://soap.sforce.com/schemas/class/BugWebServices" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns="http://schemas.xmlsoap.org/wsdl/" xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/" xmlns:tns="http://soap.sforce.com/schemas/class/BugWebServices">
 <types>
  <xsd:schema elementFormDefault="qualified" targetNamespace="http://soap.sforce.com/schemas/class/BugWebServices">
   <xsd:element name="DebuggingInfo">
    <xsd:complexType>
     <xsd:sequence>
      <xsd:element name="debugLog" type="xsd:string"/>
     </xsd:sequence>
    </xsd:complexType>
   </xsd:element>
   <xsd:simpleType name="ID">
    <xsd:restriction base="xsd:string">
     <xsd:length value="18"/>
     <xsd:pattern value="[a-zA-Z0-9]{18}"/>
    </xsd:restriction>
   </xsd:simpleType>
   <xsd:simpleType name="LogCategory">
    <xsd:restriction base="xsd:string">
     <xsd:enumeration value="Db"/>
     <xsd:enumeration value="Workflow"/>
     <xsd:enumeration value="Validation"/>
     <xsd:enumeration value="Callout"/>
     <xsd:enumeration value="Apex_code"/>
     <xsd:enumeration value="Apex_profiling"/>
     <xsd:enumeration value="All"/>
    </xsd:restriction>
   </xsd:simpleType>
   <xsd:simpleType name="LogCategoryLevel">
    <xsd:restriction base="xsd:string">
     <xsd:enumeration value="Internal"/>
     <xsd:enumeration value="Finest"/>
     <xsd:enumeration value="Finer"/>
     <xsd:enumeration value="Fine"/>
     <xsd:enumeration value="Debug"/>
     <xsd:enumeration value="Info"/>
     <xsd:enumeration value="Warn"/>
     <xsd:enumeration value="Error"/>
    </xsd:restriction>
   </xsd:simpleType>
   <xsd:complexType name="LogInfo">
    <xsd:sequence>
     <xsd:element name="category" type="tns:LogCategory"/>
     <xsd:element name="level" type="tns:LogCategoryLevel"/>
    </xsd:sequence>
   </xsd:complexType>
   <xsd:simpleType name="LogType">
    <xsd:restriction base="xsd:string">
     <xsd:enumeration value="None"/>
     <xsd:enumeration value="Debugonly"/>
     <xsd:enumeration value="Db"/>
     <xsd:enumeration value="Profiling"/>
     <xsd:enumeration value="Callout"/>
     <xsd:enumeration value="Detail"/>
    </xsd:restriction>
   </xsd:simpleType>
   <xsd:element name="DebuggingHeader">
    <xsd:complexType>
     <xsd:sequence>
      <xsd:element name="categories" minOccurs="0" maxOccurs="unbounded" type="tns:LogInfo"/>
      <xsd:element name="debugLevel" type="tns:LogType"/>
     </xsd:sequence>
    </xsd:complexType>
   </xsd:element>
   <xsd:element name="CallOptions">
    <xsd:complexType>
     <xsd:sequence>
      <xsd:element name="client" type="xsd:string"/>
     </xsd:sequence>
    </xsd:complexType>
   </xsd:element>
   <xsd:element name="SessionHeader">
    <xsd:complexType>
     <xsd:sequence>
      <xsd:element name="sessionId" type="xsd:string"/>
     </xsd:sequence>
    </xsd:complexType>
   </xsd:element>
   <xsd:element name="AllowFieldTruncationHeader">
    <xsd:complexType>
     <xsd:sequence>
      <xsd:element name="allowFieldTruncation" type="xsd:boolean"/>
     </xsd:sequence>
    </xsd:complexType>
   </xsd:element>
   <xsd:element name="updateBug">
    <xsd:complexType>
     <xsd:sequence>
      <xsd:element name="bugGetString" type="xsd:string" nillable="true"/>
     </xsd:sequence>
    </xsd:complexType>
   </xsd:element>
   <xsd:element name="updateBugResponse">
    <xsd:complexType>
     <xsd:sequence>
      <xsd:element name="result" type="xsd:boolean" nillable="true"/>
     </xsd:sequence>
    </xsd:complexType>
   </xsd:element>
   <xsd:element name="upsertComments">
    <xsd:complexType>
     <xsd:sequence>
      <xsd:element name="bugCommentsString" type="xsd:string" nillable="true"/>
     </xsd:sequence>
    </xsd:complexType>
   </xsd:element>
   <xsd:element name="upsertCommentsResponse">
    <xsd:complexType>
     <xsd:sequence>
      <xsd:element name="result" type="xsd:boolean" nillable="true"/>
     </xsd:sequence>
    </xsd:complexType>
   </xsd:element>
  </xsd:schema>
 </types>
 <!-- Message for the header parts -->
 <message name="Header">
  <part name="AllowFieldTruncationHeader" element="tns:AllowFieldTruncationHeader"/>
  <part name="CallOptions" element="tns:CallOptions"/>
  <part name="DebuggingHeader" element="tns:DebuggingHeader"/>
  <part name="DebuggingInfo" element="tns:DebuggingInfo"/>
  <part name="SessionHeader" element="tns:SessionHeader"/>
 </message>
 <!-- Operation Messages -->
 <message name="updateBugRequest">
  <part element="tns:updateBug" name="parameters"/>
 </message>
 <message name="updateBugResponse">
  <part element="tns:updateBugResponse" name="parameters"/>
 </message>
 <message name="upsertCommentsRequest">
  <part element="tns:upsertComments" name="parameters"/>
 </message>
 <message name="upsertCommentsResponse">
  <part element="tns:upsertCommentsResponse" name="parameters"/>
 </message>
 <portType name="BugWebServicesPortType">
  <operation name="updateBug">
   <input message="tns:updateBugRequest"/>
   <output message="tns:updateBugResponse"/>
  </operation>
  <operation name="upsertComments">
   <input message="tns:upsertCommentsRequest"/>
   <output message="tns:upsertCommentsResponse"/>
  </operation>
 </portType>
 <binding name="BugWebServicesBinding" type="tns:BugWebServicesPortType">
  <soap:binding style="document" transport="http://schemas.xmlsoap.org/soap/http"/>
  <operation name="updateBug">
   <soap:operation soapAction=""/>
   <input>
    <soap:header use="literal" part="SessionHeader" message="tns:Header"/>
    <soap:header use="literal" part="CallOptions" message="tns:Header"/>
    <soap:header use="literal" part="DebuggingHeader" message="tns:Header"/>
    <soap:header use="literal" part="AllowFieldTruncationHeader" message="tns:Header"/>
    <soap:body use="literal" parts="parameters"/>
   </input>
   <output>
    <soap:header use="literal" part="DebuggingInfo" message="tns:Header"/>
    <soap:body use="literal"/>
   </output>
  </operation>
  <operation name="upsertComments">
   <soap:operation soapAction=""/>
   <input>
    <soap:header use="literal" part="SessionHeader" message="tns:Header"/>
    <soap:header use="literal" part="CallOptions" message="tns:Header"/>
    <soap:header use="literal" part="DebuggingHeader" message="tns:Header"/>
    <soap:header use="literal" part="AllowFieldTruncationHeader" message="tns:Header"/>
    <soap:body use="literal" parts="parameters"/>
   </input>
   <output>
    <soap:header use="literal" part="DebuggingInfo" message="tns:Header"/>
    <soap:body use="literal"/>
   </output>
  </operation>
 </binding>
 <service name="BugWebServicesService">
  <documentation></documentation>
  <port binding="tns:BugWebServicesBinding" name="BugWebServices">
   <soap:address location="https://__SFDC_SERVER__/services/Soap/class/BugWebServices"/>
  </port>
 </service>
</definitions>
