# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RuleEngine::WebService::RuleGroup;

use 5.10.1;
use strict;
use warnings;
use base qw(Bugzilla::WebService);
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util qw(trim diff_arrays);
use Bugzilla::WebService::Util qw(validate);

use Bugzilla::Extension::RuleEngine::WebService::Util qw(rule_group_to_hash);

use Bugzilla::Extension::RuleEngine::RuleGroup;

use constant PUBLIC_METHODS => qw(
  create
  get
  list
  update
);

sub list {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);
  my @rule_groups
    = Bugzilla::Extension::RuleEngine::RuleGroup->all_user_can_view($user);

  my $data
    = {rule_groups => [map { rule_group_to_hash($self, $_) } @rule_groups]};

  return $data;
}

sub get {
  my ($self, $params) = validate(@_, 'ids', 'names');

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  if (!(defined $params->{ids} || defined $params->{names})) {
    ThrowCodeError('params_required',
      {function => 'RuleGroup.get', params => ['ids', 'names']});
  }

  my $rule_group_ids   = $params->{ids}   // [];
  my $rule_group_names = $params->{names} // [];

  my @rule_groups;

  foreach
    my $rule_group (map { Bugzilla::Extension::RuleEngine::RuleGroup->new($_) }
    @$rule_group_ids)
  {
    push(@rule_groups, $rule_group) if ($rule_group->user_can_view($user));
  }

  foreach my $rule_group (
    map { Bugzilla::Extension::RuleEngine::RuleGroup->new({name => $_}) }
    @$rule_group_names)
  {
    push(@rule_groups, $rule_group) if ($rule_group->user_can_view($user));
  }

  my $data
    = {rule_groups => [map { rule_group_to_hash($self, $_) } @rule_groups]};

  return $data;
}

sub create {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  unless ($user->can_admin_any_product()) {
    ThrowUserError("auth_failure",
      {group => 'editcomponents', action => 'edit', object => 'rule_engine'});
  }

  my @mand_args = qw/name description/;

  foreach my $field (@mand_args) {
    unless (defined($params->{$field})) {
      ThrowCodeError('params_required',
        {function => 'RuleGroup.create', params => [$field]});
    }
  }

  my $name        = delete $params->{name};
  my $description = delete $params->{description};
  my $isactive    = delete $params->{isactive} // 1;
  my $user_id     = delete $params->{user};
  my $maintainers = delete $params->{maintainers};

  my $view_groups = delete $params->{'view_groups'};
  my $edit_groups = delete $params->{'edit_groups'};

  my @invalid_params = keys(%$params);
  ThrowUserError('invalid_parameters',
    {function => 'RuleGroup.create', params => \@invalid_params})
    if (@invalid_params);

  my @view_group_obj = map { Bugzilla::Group->check({name => $_}) } @$view_groups;
  my @edit_group_obj = map { Bugzilla::Group->check({name => $_}) } @$edit_groups;

  Bugzilla->dbh->bz_start_transaction;

  my %args = (name => $name, description => $description, isactive => $isactive,);

  my $rule_group = Bugzilla::Extension::RuleEngine::RuleGroup->create(\%args);

  $rule_group->set_user_id($user_id)         if ($user_id);
  $rule_group->set_maintainers($maintainers) if ($maintainers);

  if (scalar @view_group_obj) {
    $rule_group->set_view_groups(\@view_group_obj);
  }
  if (scalar @edit_group_obj) {
    $rule_group->set_edit_groups(\@edit_group_obj);
  }

  $rule_group->update();

  Bugzilla->dbh->bz_commit_transaction;

  return ({rule_group => rule_group_to_hash($self, $rule_group)});
}

sub update {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $rule_groups
    = delete $params->{rule_groups}
    || ThrowCodeError('params_required',
    {function => 'RuleGroup.update', params => ['rule_groups']});

  my @invalid_params = keys(%$params);
  ThrowUserError('invalid_parameters',
    {function => 'RuleGroup.update', params => \@invalid_params})
    if (@invalid_params);

  Bugzilla->dbh->bz_start_transaction;

  my @changes;

  foreach my $r_group (@$rule_groups) {

    my $name = delete $r_group->{name};
    my $id   = delete $r_group->{id};

    my $rule_group;
    if ($id) {
      $rule_group = Bugzilla::Extension::RuleEngine::RuleGroup->check({id => $id});
    }
    elsif ($name) {
      $rule_group
        = Bugzilla::Extension::RuleEngine::RuleGroup->check({name => $name});
    }
    else {
      ThrowCodeError('params_required',
        {function => 'RuleGroup.update', params => ['id', 'name']});
    }

    my $description = delete $r_group->{description};

    my $change_isactive = defined $r_group->{'isactive'};
    my $isactive        = delete $r_group->{isactive};
    my $user_id         = delete $r_group->{user};
    my $maintainers     = delete $r_group->{maintainers};

    my $change_v_groups = defined $r_group->{'view_groups'};
    my $change_e_groups = defined $r_group->{'edit_groups'};
    my $view_groups     = delete $r_group->{'view_groups'};
    my $edit_groups     = delete $r_group->{'edit_groups'};

    @invalid_params = keys(%$r_group);
    ThrowUserError('invalid_parameters',
      {function => 'RuleGroup.update', params => \@invalid_params})
      if (@invalid_params);

    my @view_group_obj = map { Bugzilla::Group->check({name => $_}) } @$view_groups;
    my @edit_group_obj = map { Bugzilla::Group->check({name => $_}) } @$edit_groups;

    $rule_group->set_name($name)               if ($name);
    $rule_group->set_description($description) if ($description);
    $rule_group->set_is_active($isactive)      if ($change_isactive);
    $rule_group->set_user_id($user_id)         if ($user_id);

    my $changes = $rule_group->update();

    if ($change_v_groups) {
      $rule_group->set_view_groups(\@view_group_obj);

      my ($removed, $added)
        = diff_arrays($rule_group->{_old_view_groups}, $rule_group->{view_groups});

      if (@$removed || @$added) {
        $changes->{view_groups} = [[sort map { $_->{name} } @{$removed}],
          [sort map { $_->{name} } @{$added}]];
      }
    }

    if ($change_e_groups) {
      $rule_group->set_edit_groups(\@edit_group_obj);

      my ($removed, $added)
        = diff_arrays($rule_group->{_old_edit_groups}, $rule_group->{edit_groups});

      if (@$removed || @$added) {
        $changes->{edit_groups} = [[sort map { $_->{name} } @{$removed}],
          [sort map { $_->{name} } @{$added}]];
      }
    }

    if ($maintainers) {
      $rule_group->set_maintainers($maintainers);

      my ($removed, $added)
        = diff_arrays($rule_group->{'_old_maintainers'}, $rule_group->{maintainers});

      if (@$removed || @$added) {
        $changes->{maintainers}
          = [[sort map { $_->login } @{$removed}], [sort map { $_->login } @{$added}]];
      }
    }

    push(@changes, $changes);
  }

  Bugzilla->dbh->bz_commit_transaction;

  return ({changes => \@changes});
}

1;

__END__


=head1 NAME

Bugzilla::Extension::RuleEngine::WebService::RuleGroup - The RuleEngine RuleGroup
API

=head1 DESCRIPTION

This API allows you to list, get, create and update rule groups.

=head1 METHODS

=head2 list

=over

=item B<Description>

Returns information about all the rule groups a user can view.

=item B<Params>

none

=item B<Returns>

Returns C<rule_groups> an array of hashes containing rule groups.

=back

=head2 get

=over

=item B<Description>

Returns information about the selected rule groups.

=item B<Params>

=over

=item B<names>

An array of names to get information on.

=item B<ids>

An array of IDs to get information on.

=back

=item B<Returns>

Returns C<rule_groups> an array of hashes containing the selected rule groups.

=back

=head2 create

=over

=item B<Description>

Create a new rule group.

=item B<Params>

=over

=item B<name>

C<string>  The name of the rule group.

=item B<description>

C<string> The description of the rule group.

=item B<isactive>

C<boolean> True if the rule group is active.  False if inactive.

=item B<user>

C<string> The user name to run rules in this rule group as. Must be a valid
Bugzilla account.

=item B<maintainers>

C<array-of-string> An array of Bugzilla user names to notify of rule changes and
kill switches.

=item B<view_groups>

C<array-of-string> An array of Bugzilla group names who can view rules in this
rule group.

=item B<edit_groups>

C<array-of-string>  An array of Bugzilla group names who can edit rules in this
rule group.

=back

=item B<Returns>

Returns C<rule_group> a hash containing the new rule group.

=back

=head2 update

=over

=item B<Description>

Update an existing rule group.

=item B<Params>

=over

=item B<rule_groups>

C<array-of-hashes> An array of hashes, each item containing a separate rule group
to edit.
Each hash may contain:

=over

=item B<id> I<optional>

C<int> The ID of the rule group. If this is supplied then the name of the rule
group can be changed.

=item B<name> I<optional>

C<string> The name of the rule group. If ID is supplied the name will be
updated.

B<Note> Either name or id must be supplied.

=item B<description> I<optional>

C<string> The description of the rule group.

=item B<isactive> I<optional>

C<boolean> True if the rule group is active.  False if inactive.

=item B<user> I<optional>

C<string> The user name to run rules in this rule group as. Must be a valid
Bugzilla account.

=item B<maintainers> I<optional>

C<array-of-string> An array os Bugzilla user names to notify of rule changes and
kill switches.

=item B<view_groups> I<optional>

C<array-of-string> An array os Bugzilla group names who can view rules in this
rule group.

=item B<edit_groups> I<optional>

C<array-of-string>  An array os Bugzilla group names who can edit rules in this
rule group.

=back

=back

=item B<Returns>

Returns C<changes> an array of hashes.  Each hash describes the changes to one
rule group. Only fields that were changed will be returned.

Each field will be returned as an array. Element 0 will be the old value that was
replaced. Element 1 will be the new value for that field.

=back
