# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::SAML2Auth::Util;

use 5.10.1;
use strict;
use warnings;

use Net::SAML2;

use Bugzilla::Error;
use Bugzilla::Extension::SAML2Auth::IDP;
use Bugzilla::Util qw(correct_urlbase template_var trick_taint);

use parent qw(Exporter);
our @EXPORT = qw(
  sp
  idp
);

# This file can be loaded by your extension via
# "use Bugzilla::Extension::SAML2Auth::Util". You can put functions
# used by your extension in here. (Make sure you also list them in
# @EXPORT.)

sub sp {
  return (Bugzilla->request_cache->{saml2auth_sp})
    if (Bugzilla->request_cache->{saml2auth_sp});
  my $sp = Net::SAML2::SP->new(
    id               => correct_urlbase() . 'saml2_metadata.cgi',
    url              => correct_urlbase(),
    cert             => Bugzilla->params->{saml2auth_sp_crt},
    key              => Bugzilla->params->{saml2auth_sp_key},
    cacert           => Bugzilla->params->{saml2auth_sp_cacert},
    org_name         => template_var('terms')->{Bugzilla},
    org_display_name => template_var('terms')->{Bugzilla},
    org_contact      => correct_urlbase(),
  );

  Bugzilla->request_cache->{saml2auth_sp} = $sp;

  return (Bugzilla->request_cache->{saml2auth_sp});
}

sub idp {
  my $idp_name = shift @_;

  ThrowUserError('saml_no_idp') unless ($idp_name);

  trick_taint($idp_name);

  return (Bugzilla->request_cache->{"saml2auth_idp_$idp_name"})
    if (Bugzilla->request_cache->{"saml2auth_idp_$idp_name"});

  my $idp;
  eval { $idp = Bugzilla::Extension::SAML2Auth::IDP->new({name => $idp_name}); };
  ThrowUserError('saml_no_such_idp', {name => $idp_name}) unless ($idp);

  if (($idp->metadata eq "") or ($idp->expired)) {
    my $uri = URI->new($idp->url);
    my $ua  = LWP::UserAgent->new;

    my $req = HTTP::Request->new(GET => $uri);
    my $res = $ua->request($req);

    unless ($res->is_success) {
      ThrowUserError('saml_idp_fetch_error', {msg => $res->status_line});
    }

    my $metadata = $res->content;

    $idp->set_metadata($metadata);
    $metadata =~ m{validUntil="([^"]*)"};
    my $expires
      = $1
      ? $1
      : Bugzilla->dbh->selectrow_array(
      q{select date_trunc('day', NOW() + interval '1 month')});
    $idp->set_expires($expires);
    $idp->update();
  }

  my $sidp;
  eval {
    $sidp
      = Net::SAML2::IdP->new_from_xml(xml => $idp->metadata,
      cacert => $idp->cacert,);
  };
  if ($@) {
    my $msg = $@;
    ThrowUserError('saml_idp_metadata_error', {msg => $msg});
  }

  Bugzilla->request_cache->{"saml2auth_idp_$idp_name"} = $sidp;
  return (Bugzilla->request_cache->{"saml2auth_idp_$idp_name"});
}

1;

__END__

=head1 Description

Bugzilla::Extension::SAML2Auth::Util - A module for suppliy utility functions for SAML2 AUthentication.


=head1 Methods

=over 4

=item sp

Returns a handle to a L<Net::SAML2::SP> object for this servers SP data.

=item idp

Returns a handle to a L<Net::SAML2::IdP> object for the named IDP.

=back

