# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::SelectizeJS::WebService::Search;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::WebService);
use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::Util;
use Bugzilla::WebService::Util qw(validate);
use Data::Dumper;
use Bugzilla::Product;
use Bugzilla::User;
use Bugzilla::Extension::SubComponents::RHSubComponent;

use List::MoreUtils qw(any none);

use constant PUBLIC_METHODS => qw(
  list_products
  list_components
  is_prod_multi
  list_subcomponents
  list_roles
);

# function to return product information when passed either,
# product ids as used on edit page select, term to search on, wheter or not to return the product descriptions
# exmaple of how it will be called:
# $call = $rpc->call( 'SelectizeJS.list_products',
#                          { term => 'ker'
#                          }
#                   );

sub list_products {
  my ($self, $params) = validate(@_, 'term', 'classification', 'disabled');

  Bugzilla->switch_to_shadow_db();
  my $user = Bugzilla->login(LOGIN_OPTIONAL);

  # filter out products the user can't see.
  my @prods = @{Bugzilla->user->get_accessible_products};
  if (defined $params->{classification}) {
    my @clas_prods;

    # Only prune products if classifications are passed in, else return them all.
    if (@{$params->{classification}}) {
      foreach my $clas (@{$params->{classification}}) {
        push @clas_prods, grep { $_->classification->name eq $clas } @prods;
      }
      @prods = sort { $a->name cmp $b->name } @clas_prods;
    }
  }

  if (!defined $params->{disabled} || !$params->{disabled}[0]) {
    my @active_prods;
    push @active_prods, grep { $_->is_active == 1 } @prods;

    @prods = @active_prods;
  }

  my @results = map +{
    id             => 'v' . $_->id . '_product',
    value          => $_->name,
    text           => $_->name,
    title          => $_->description,
    classification => $_->classification->name
  }, @prods;

  return \@results;
}

# function to return component information when passed either,
# product ids (as used on edit page select) or a single product name, term to search on, whether or not to return the component descriptions
# exmaple of how it will be called:
# $call = $rpc->call( 'SelectizeJS.list_components',
#                          {product => 'v351_product', term => 'ker', descrs => 1
#                          }
#                   );
# Use components to search one or more components by exact name.
# Exmaple of how it will be called:
# $call = $rpc->call( 'SelectizeJS.list_components',
#                          {product => 'Bugzilla', components => [comp_1, comp_2]
#                          }
#                   );
# NOTE: term and components are mutually exclusive.
#
sub list_components {
  my ($self, $params)
    = validate(@_, 'product', 'term', 'descrs', 'disabled', 'components');

  Bugzilla->switch_to_shadow_db();
  my $user = Bugzilla->login(LOGIN_OPTIONAL);

  # filter out products the user can't see.
  my %accessible_product_ids
    = map { $_->id => 1 } @{Bugzilla->user->get_accessible_products};

  my %seen;
  my @results;
  my $dbh = Bugzilla->dbh();
  my @prod_ids;
  my $id_str     = '';
  my $product    = $params->{product}[0] || die "No product";
  my $term       = $params->{term}[0];
  my $disabled   = $params->{disabled}[0] || 0;
  my @components = @{$params->{components} || []};
  map { trick_taint($_) } @components;

  if (!@components && !$term) {
    die "No components or term";
  }
  elsif (@components && $term) {
    die "Components and term are mutually exclusive";
  }

  my @binding;
  my $num_prods = 0;
  foreach my $prod (split(/,/, $product)) {
    my $prod_id;
    if ($prod =~ /^v?(\d+)(?:_product)?$/) {
      $prod_id = $1;
    }
    else {
      my $prod = Bugzilla::Product->new({name => $product});
      $prod_id = $prod->id;
    }

    unless ($accessible_product_ids{$prod_id}) {
      Bugzilla->logger->info("User does not have access to product $prod_id\n");
      next;
    }

    push(@prod_ids, $prod_id);
    $id_str .= '?,';
  }
  push @binding, @prod_ids;
  return if ($id_str eq '');

  chop($id_str);

  my $query = "SELECT components.id, components.name";
  $query .= ", components.description" if ($params->{descrs});
  $query .= " FROM components WHERE product_id in ($id_str)";

  my @order_by = ('components.name');
  if ($term) {
    $query
      .= " AND "
      . $dbh->sql_iposition($dbh->quote($term), 'components.name') . ' > 0'
      if ($term ne "..");
    unshift @order_by, $dbh->sql_iposition($dbh->quote($term), 'components.name');
  }
  else {
    $query .= " AND components.name IN (" . join(', ', ('?') x @components) . ")";
    push @binding, @components;
  }

  $query .= " AND components.isactive = 1" unless ($disabled);
  $query .= " ORDER BY " . join(', ', @order_by);

  my $component_list = $dbh->selectall_arrayref($query, undef, @binding);

  foreach my $row (@{$component_list}) {
    my ($comp_id, $comp_name, $descr) = @{$row};
    unless (defined($seen{$comp_name})) {
      my %res = (value => $comp_name, text => $comp_name, id => $comp_id);
      $res{title} = $descr if ($params->{descrs});
      push(@results, \%res);
      $seen{$comp_name} = 1;
    }
  }

  return \@results;
}

# function to return whether or not a product allows multiple components on a bug,
# product ids as used on edit page select
# exmaple of how it will be called:
# $call = $rpc->call( 'SelectizeJS.list_components',
#                          {prduct => 'v351_product'
#                          }

sub is_prod_multi {
  my ($self, $params) = validate(@_, 'product');

  Bugzilla->switch_to_shadow_db();
  my $user = Bugzilla->login(LOGIN_OPTIONAL);

  # filter out products the user can't see.
  my %accessible_product_ids
    = map { $_->id => 1 } @{Bugzilla->user->get_accessible_products};

  my $product = $params->{product}[0] || die "No product";
  $product =~ /^v?(\d+)(?:_product)?$/ || die "Bad product id: $product";
  my $prod_id = $1;

  return unless ($accessible_product_ids{$prod_id});

  my $prod = Bugzilla::Product->new({id => int($prod_id)});
  return {ismulti => $prod->multiple_components};
}

# function to return subcomponent information when passed either,
# product id as used on edit page select, component names,
# use select to mark the sub-components as selected
# exmaple of how it will be called:
# $call = $rpc->call( 'SelectizeJS.list_subcomponents',
#                          {product => 'v351_product', components = 'kernel',
#                          }
#                   );
# use_name is for the search page which uses the full tree of sub component names to search on.

sub list_subcomponents {
  my ($self, $params)
    = validate(@_, 'product', 'component', 'descrs', 'select', 'disabled',
    'use_name');

  Bugzilla->switch_to_shadow_db();
  my $user = Bugzilla->login(LOGIN_OPTIONAL);

  # filter out products the user can't see.
  my %accessible_product_ids
    = map { $_->id => 1 } @{Bugzilla->user->get_accessible_products};

  my @results;
  my $dbh       = Bugzilla->dbh();
  my $id_str    = '';
  my $product   = $params->{product}[0] || die "No product";
  my $component = $params->{component}[0] || die 'No component';
  my @select    = @{$params->{select} || []};
  map { trick_taint($_) } @select;
  my $disabled = $params->{disabled}[0] || 0;
  my $use_name = $params->{use_name}[0] || 0;

  my @prod_ids;
  foreach my $prod (split(/,/, $product)) {
    $prod =~ /^v?(\d+)(?:_product)?$/ || die "Bad product id: $prod";
    my $prod_id = $1;
    unless ($accessible_product_ids{$prod_id}) {
      Bugzilla->logger->info("User does not have access to product $prod_id\n");
      next;
    }

    push(@prod_ids, $prod_id);
    $id_str .= '?,';
  }
  die 'No accessible products supplied' if ($id_str eq '');
  chop($id_str);


  my @components = @{$params->{component}};
  map { trick_taint($_) } @components;

  my $count     = @{$params->{component}};
  my $param_str = '?,' x $count;
  chop($param_str);

  my $query = <<EOS;
  SELECT rh_sub_components.id, rh_sub_components.name, rh_sub_components.description, components.id,components.product_id
    FROM rh_sub_components left join components on rh_sub_components.component_id = components.id
   WHERE components.name in ($param_str)
    AND  components.product_id in ($id_str)
EOS

  $query .= " AND rh_sub_components.isactive = 1 " unless ($disabled);

  $query .= <<EOS;
    AND  level = 0
ORDER BY rh_sub_components.sortkey, rh_sub_components.name
EOS

  trick_taint($component);

  my $subcomponents
    = $dbh->selectall_arrayref($query, undef, @components, @prod_ids);

  foreach my $row (@{$subcomponents}) {
    my ($sub_comp_id, $sub_comp_name, $sub_descr, $comp_id, $prod_id) = @{$row};
    my $value = $sub_comp_id;
    $value = $sub_comp_name if ($use_name);

    my %res = (
      value   => $value,
      text    => $sub_comp_name,
      title   => $sub_descr,
      comp_id => $comp_id
    );
    $res{selected} = 1 if (any { $_ eq $sub_comp_name } @select);
    push(@results, \%res);
    _get_subs(\@results, $component, $prod_id, $sub_comp_id, $use_name, $value,
      \@select);
  }

  return \@results;
}

sub _get_subs {
  my ($results, $component, $prod_id, $parent_id, $use_name, $label, $select)
    = @_;

  my $dbh   = Bugzilla->dbh();
  my $query = <<'EOS';
  SELECT rh_sub_components.id, rh_sub_components.name, rh_sub_components.description, components.id, level
    FROM rh_sub_components left join components on rh_sub_components.component_id = components.id
   WHERE components.name = ?
    AND  components.product_id = ?
    AND  rh_sub_components.isactive = 1
    AND  parent_id = ?
ORDER BY rh_sub_components.sortkey, rh_sub_components.name
EOS

  my $subcomponents
    = $dbh->selectall_arrayref($query, undef, $component, $prod_id, $parent_id);

  foreach my $row (@{$subcomponents}) {
    my ($sub_comp_id, $sub_comp_name, $sub_descr, $comp_id, $level) = @{$row};
    my $value = $sub_comp_id;
    $value = "$label $sub_comp_name" if ($use_name);

    my %res = (
      value   => $value,
      text    => $sub_comp_name,
      title   => $sub_descr,
      comp_id => $comp_id,
      level   => $level
    );
    $res{selected} = 1 if (any { $_ eq $sub_comp_name } @$select);
    push(@$results, \%res);
    _get_subs($results, $component, $prod_id, $sub_comp_id, $use_name, $value,
      $select);
  }

  return;
}

# function to return role information when passed
# product name, component name, and optionally subcomponent id
# exmaple of how it will be called:
# $call = $rpc->call( 'SelectizeJS.list_roles',
#                          {product => 'Bugzilla', component => 'Administration'
#                          }
#                   );
sub list_roles {
  my ($self, $params) = validate(@_, 'product', 'component', 'subcomponent');

  Bugzilla->switch_to_shadow_db();
  my $user = Bugzilla->login(LOGIN_OPTIONAL);

  # filter out products the user can't see.
  my %accessible_product_ids
    = map { $_->id => 1 } @{Bugzilla->user->get_accessible_products};

  my @results;
  my $dbh = Bugzilla->dbh();
  my @prod_ids;
  my $id_str       = '';
  my $product      = $params->{product}[0] || die "No product";
  my $component    = $params->{component}[0] || die 'No component';
  my $subcomponent = $params->{subcomponent}[0];
  my %res;

  my $prod = new Bugzilla::Product({name => $product});
  return () unless ($accessible_product_ids{$prod->id});

  my $comp = Bugzilla::Component->new({product => $prod, name => $component});

  unless ($comp) {
    Bugzilla->logger->info(
      "No component found for product $product, component $component\n");
    return ();
  }

  $res{assignee} = _user2hash($comp->default_assignee)
    if ($comp->default_assignee);
  $res{qa_contact} = _user2hash($comp->default_qa_contact)
    if ($comp->default_qa_contact);
  $res{docs_contact} = _user2hash($comp->default_docs_contact)
    if ($comp->default_docs_contact);

  $res{initial_cc} = join(', ', map { $_->{login_name} } @{$comp->initial_cc})
    if ($comp->initial_cc);
  $res{flags} = $comp->flag_types();

  if ($subcomponent) {
    $subcomponent =~ /^([0-9]+)$/ || die "Bad subcomponent id '$subcomponent'";
    $subcomponent = $1;

    my $sc = Bugzilla::Extension::SubComponents::RHSubComponent->new($subcomponent);
    unless ($sc) {
      Bugzilla->logger->info(
        "No subcomponent found for product $product, component $component, subcomponent $subcomponent\n"
      );
      return ();
    }
    $res{assignee} = $sc->default_assignee->{login_name} if ($sc->default_assignee);
    $res{qa_contact} = $sc->default_qa_contact->{login_name}
      if ($sc->default_qa_contact);
    $res{docs_contact} = $sc->default_docs_contact->{login_name}
      if ($sc->default_docs_contact);
    $res{initial_cc} = join(', ', map { $_->{login_name} } @{$sc->initial_cc})
      if ($sc->initial_cc);
  }

  push(@results, \%res);

  return \@results;
}

sub _user2hash {
  my $the_user = shift;

  my %user_data
    = (login => $the_user->{login_name}, insider => $the_user->is_insider,);

  return \%user_data;
}

1;
