#!/usr/bin/perl -wT
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

## REDHAT EXTENSION 1104055

use strict;
use warnings;
use lib qw(. lib);
use 5.10.1;

use Bugzilla;
use Bugzilla::Error;
use Bugzilla::Mailer;
use Bugzilla::Util;

use DateTime;
use List::MoreUtils qw(uniq);

# Create some variables that we'll need
my $dbh          = Bugzilla->dbh;
my $today        = DateTime->today->ymd;
my $yesterday    = DateTime->today->subtract(days => 1)->ymd;
my $default_lang = Bugzilla::User->new()->setting('lang');

# The email client will display the Date: header in the desired timezone,
# so we can always use UTC here.
my $timestamp = $dbh->selectrow_array('SELECT LOCALTIMESTAMP(0)');
$timestamp = format_time($timestamp, '%a, %d %b %Y %T %z', 'UTC');

# What things we will be checking
my @checks = (
  {
    product_re   => qr{^Red Hat Enterprise Linux [567]$},
    component_re => qr{^kernel$},
    email_to     => ['rcain@redhat.com', 'caiqian@redhat.com', 'lwang@redhat.com']
  },
  {product_re => qr{^Bugzilla$}, email_to => ['sgreen@redhat.com']}
);

# Translate the field into plain English
my %word = (
  initialowner       => 'Default Assignee',
  initialqacontact   => 'Default QA Contact',
  initialdocscontact => 'Default Docs Contact',
  initialcc          => 'Default CC',
);

# Get all sub component changes yesterday
my $query = q{
    SELECT p.name, c.name, s.full_name, a.field, a.added, a.removed,
        u.login_name, a.at_time
    FROM audit_log a
        JOIN rh_sub_components s ON (a.object_id = s.id)
        JOIN components c ON (s.component_id = c.id)
        JOIN products p ON (c.product_id = p.id)
        JOIN profiles u ON (a.user_id = u.userid)
    WHERE a.class = 'Bugzilla::Extension::SubComponents::RHSubComponent'
        AND a.field IN ('initialcc', 'initialdocscontact', 'initialowner', 'initialqacontact')
        AND a.at_time >= ? AND a.at_time < ?
    ORDER BY a.at_time
};

my $changes = $dbh->selectall_arrayref($query, undef, $yesterday, $today);

foreach (@$changes) {
  my ($product, $component, $sub_comp, $field, $added, $removed, $who, $when)
    = @$_;

  my $what = $word{$field};

  # Do we want to be notified about this change?
  foreach my $check (@checks) {
    if ($check->{product_re} && $product !~ $check->{product_re}) {
      next;
    }
    if ($check->{component_re} && $component !~ $check->{component_re}) {
      next;
    }

    # So we are going to send out an e-mail. Lets get the e-mail addresses
    # of the users added and removed (if any)
    my @added_emails
      = (map { Bugzilla::User->new($_)->login } split(/\D+/, ($added // '')));
    my @removed_emails
      = (map { Bugzilla::User->new($_)->login } split(/\D+/, ($removed // '')));

    # Turn this information to a hash with e-mail as the key and the
    # reason as an arrayref value
    my %reason = _rh_subcomp_reason(
      "You made the change",
      [$who],
      "You were the $what",
      \@removed_emails,
      "You are the $what",
      \@added_emails,
      "You requested to be notified about changes ",
      ($check->{email_to} // [])
    );

    foreach my $to (keys %reason) {
      my $vars = {
        to             => $to,
        reasons        => $reason{$to},
        date           => $timestamp,         # e-mail date
        product_name   => $product,
        component_name => $component,
        sub_comp_name  => $sub_comp,
        what           => $what,
        added          => \@added_emails,
        removed        => \@removed_emails,
        who            => $who,
        when           => $when,              # change made date
      };

      my $template = Bugzilla->template_inner($default_lang);
      my $message;
      $template->process("email/sub_comp_change.txt.tmpl", $vars, \$message)
        || ThrowTemplateError($template->error());

      MessageToMTA($message);
    }
  }
}

sub _rh_subcomp_reason {
  my %reason = ();
  my @input  = @_;

  while (@input) {
    my $text   = shift @input;
    my $logins = shift @input;
    foreach my $login (@$logins) {
      push @{$reason{$login}}, $text;
    }
  }

  return %reason;
}

1;
